;**********************************************************
;***  	   Universal Light Controller Version 1.0       ***
;**********************************************************
;supported PICs: 12F675, 16F630, 16F636, 16F648A
;
;History:
;1.0 initial Version
;1.0 added support of PIC16F648A
;1.1 Bugfix: Macros JSWXON and JSWXOFF ignored switch number    
;1.1 added RC inactive switch    
;__________________________________________________________
;!!!!!!!!!   Limitations of this implementation:    !!!!!!!
;interpolation (Smooth-mode) is interrupted by any control
;instruction (Op-code >= 0xC0).
;RNDDLY and RNDBRT must not be used when Smooth mode is on.
;__________________________________________________________

;Port usage (16F63x):
;RA3 = input: RC-channel 1
;RA0 = input: RC-channel 2
;RC0..RC5 = LED outputs (1: LED is on) 0..5
;RA4,RA5 = LED outputs 6,7

;Port usage (16F627, 16F628A or 16F648A):
;RB0/INT = input: RC-channel 1
;RB4..RB7 = LED outputs 0..3
;RA6,RA7,RA0,RA4 = LED outputs 4..7

;Port usage (12Fxxx):
;GP3 = input: RC impulse, channel 1
;GP5 = input: RC impulse, ch.2 (if RCCHAN =2)
;GP0..GP2, GP4, GP5: LED outputs (max. 5 channels)
	
	radix dec

	if NCHAN > 8
	error "Number of output channels out of range!"
	endif 

    ifdef __16F630
	#include <p16F630.inc>         	; processor specific variable definitions
	#define MAXRAM 0x5F
;config for 16F630:
	__config _INTRC_OSC_NOCLKOUT & _WDT_OFF & _PWRTE_ON & _MCLRE_OFF

	if CLOCK != 4
		messg "Clock frequency setting changed to 4 MHz"
		#undefine CLOCK
		#define CLOCK 4
	endif
    endif

    ifdef __16F636
	#include <p16F636.inc>         	; processor specific variable definitions
	#define MAXRAM 0x7F
;config for 16F636:
	__config _INTRC_OSC_NOCLKOUT & _WDT_OFF & _PWRTE_ON & _MCLRE_OFF
    endif

    ifdef __16F648A
	#include <p16F648A.inc>        	; processor specific variable definitions
	#define MAXRAM 0x7F
;config for 16F428:
	__config _INTRC_OSC_NOCLKOUT & _WDT_OFF & _PWRTE_ON & _MCLRE_OFF & _LVP_OFF
	if RCCHAN > 1
		error "Only one RC-input supported!"
	endif
    endif

    ifdef __12F675
	#include <p12F675.inc>         	; processor specific variable definitions
	#define MAXRAM 0x5F
;config for 12F675:
	__config _INTRC_OSC_NOCLKOUT & _WDT_OFF & _PWRTE_ON & _MCLRE_OFF
#define RAIF GPIF		;some re-definitions, cause code is written for 16Fxxx...
#define RAIE GPIE
#define PORTC GPIO
	if NCHAN > 5
		error "Number of output channels out of range!"
	else    	
		if (RCCHAN >1)&&(NCHAN>4) 
		error "Number of output channels out of range!"
		endif
    	endif 
	if CLOCK != 4
		messg "Clock frequency setting changed to 4 MHz"
		#undefine CLOCK
		#define CLOCK 4
	endif
    endif

 if (PWMTU * CLOCK) < (110 + NCHAN * 22)
	error "PWMTU must be higher value!"
 endif  

;*********************************************************************
;Macro: checks if (addr) = EEPROM address:
chkaddr	macro addr
 if high addr != 0x21
	error "Invalid jump address"
 endif
	endm

;ALIC Opcode macro definitions:

SETBRIGHT macro value,delay
 if delay>255
	error "SETBRIGHT: Delay out of range"
 endif
 if value>=0xC0
	error "SETBRIGHT: Brightness value out of range"
 endif
	de	value,delay
	endm

LIGHTOFF macro delay
 if delay>255
	error "LIGHTOFF: Delay out of range"
 endif
	de	0,delay
	endm

LIGHTON macro delay
 if delay>255
	error "LIGHTON: Delay out of range"
 endif
	de	0x80,delay
	endm

SMOOTHOFF macro
	de	0xC0
	endm

SMOOTHON macro
	de	0xC1
	endm

HALTSWOFF macro
	de	0xC2
	endm

HALTSWON macro
	de	0xC3
	endm

JSWOFF 	macro	addr
	chkaddr addr
	de	0xC4,low addr
	endm

JSWON 	macro	addr
	chkaddr addr
	de	0xC5,low addr
	endm

CLRSW 	macro
	de	0xC6
	endm

SETSW 	macro
	de	0xC7
	endm

JUMP 	macro	addr
	chkaddr	addr
	de	0xC8,low addr
	endm

RNDDLY 	macro	maxticks
	de	0xC9,maxticks
	endm

DELAY 	macro	ticks
	de	0xDC,ticks
	endm

RNDBRT	macro 	min,max,delay
 if (min >= 0xC0) || (max>=0xC0)
	error "RNDBRT: Brightness out of range"
 endif
	de	0xCA,min,max,delay
	endm

RNDJMP	macro	chance,addr
	chkaddr	addr
	de	0xCB,chance,low addr
	endm

JSWXOFF	macro	switch,addr
	chkaddr	addr
	de	0xCC,switch,low addr
	endm

JSWXON	macro	switch,addr
	chkaddr	addr
	de	0xCD,switch,low addr
	endm

CLRSWX	macro	switch
	de	0xCE,switch
	endm

SETSWX	macro	switch
	de	0xCF,switch
	endm

DEFTICKLEN macro ms
 if ms<10
	messg "DEFTICKLEN: Tick length should not be below 10"
 endif	
	de	0xD0,ms
	endm

DEFBRIGHT macro value
	de	0xD1,value
	endm

DEFSWITCH macro switch
	de	0xD2,switch
	endm

DEFVAR	macro	var
	de	0xD3,var
	endm

JVARB	macro	val,addr
	chkaddr addr
	de	0xD8,val,low addr
	endm

JVARAE	macro	val,addr
	chkaddr addr
	de	0xD9,val,low addr
	endm

JVARXB	macro	var,value,addr
	chkaddr	addr
	de	0xDA,var,value,low addr
	endm

JVARXAE macro 	var,value,addr
	chkaddr	addr
	de 	0xDB,var,value,low addr
	endm

P_RESET macro
	de	0xFF
	endm
;**************************************************************

    if COMMTICK
#define TICKTIME 10000		;time for 1 systick (=10 ms)
CTCNT equ 0
    else
#define TICKTIME 1000		;time for 1 systick (=1 ms)
CTCNT equ NCHAN
    endif
 if CLOCK == 4
;constant for loading TMR0 @ 4 MHz Clock:
#define TMR0LOAD PWMTU-3
 else
;constant for loading TMR0 @ 8 MHz Clock:
#define TMR0LOAD PWMTU*2-3
 endif 

 if RCCHAN
#define RCCTRL 1
 else
#define RCCTRL 0
 endif
 
;flag-bits:
;#define JUMPFLG 0		;jump was executed in current tick-processing
#define SKIPSMTH 1		;temporarily skip smooth processing
;#define RCFLG1 2		;valid new RC-pulse value (rcval ch.1)
;#define RCFLG2 3		;valid new RC-pulse value (rcval ch.2)
#define RCINIT 4		;RC initialization phase if 0
#define RC2ENA 5		;enable two RC inputs
#define	RCSELECT 6		;currently selected RC-channel
#define SAVERP0 7		;save RP0 (RC-interrupt)

;*********** RAM
	CBLOCK	0x20
;data that are used or altered in interrupt service routines:
pwmval:	NCHAN			;current brightness for each channel
pwmcnt: NCHAN			;time till next pwm-action (each channel) 
timesum: 2			;used to create systicks
systicks			;incremented each TICKTIME by ISR

channel				;register for currently selected channel (1..NCHAN)
y1,y2,y,x1,x2,x			;used for arithmetic (interpolation, multiplication...) 

switch				;8 control switches (1 per bit)
smooth				;8 smooth mode bits
systcnt: CTCNT			;systick countdown for each output channel
tickcnt: NCHAN			;timer: waiting for next step
pcpip:	NCHAN			;instruction pointer: current PCP program step

random				;8 bit random number

errcnt: RCCTRL			;counts invalid rc-pulses
				;lower nibble: RC1, higher nibble: RC2
rctimer: RCCTRL			;1/10 sec. timer for RC 
				;lower nibble: RC1, higher nibble: RC2
rcdetect: RCCTRL		;MSB = 1: no valid RC signal detected
rcval: RCCHAN			;current RC pulse value (0..127), RC1
rclast: RCCHAN			;last pulse value, RC1
memswtch: RCCTRL		;bit 0..3 = RC memory switch bits, toggled by:
				;bit 0: stick moved down for short time
				;bit 1: stick moved down for long time
				;bit 2: stick moved up for short time
				;bit 3: stick moved up for long time
				;(higher nibble: same for RC2
tc_100				;time downcounter @ systick-rate for 100ms
;time10				;increments each 1/10 second
	ENDC

    ifdef __16F636
	#define XVARS	0x77	;data accessable in all banks
    endif
    ifdef __16F648A
	#define XVARS	0x77	;data accessable in all banks
    endif
    ifndef XVARS
	#define XVARS		;empty expr. = CBLOCK uses current location	
    endif

	CBLOCK	XVARS
temp				;temporary data
flags				;various flags
;ee_edit:2			;address & data for edit parameter in EEPROM			

t0intcnt			;number or t0-int service calls to be executed
save_W				;for saving registers (interrupt)
save_SR
save_W2
save_SR2
	ENDC
 if save_SR2 > MAXRAM
	error "Out of RAM!"
 endif
;**************************************************************
;EEPROM:
	ORG	0x2100
entrytable macro
	local ch=0
	while ch<NCHAN
		de low entry_ch#v(ch)
ch = ch+1
	endw
	endm

	entrytable
#include ALICFILE

;**************************************************************
;Main Program:

reset	ORG	0
 if ACTIVELOW
	movlw	0x7F		;active low outputs: 1 = off
 else
	movlw	0x00		;active high outputs: 0 = off
 endif
 ifndef __12F675
	movwf	PORTA
 endif
 ifdef	__16F648A
	movwf	PORTB
 else
	movwf	PORTC		;all LEDs off
 endif
	goto	init

	ORG	4
;*********** Interrupt service routine
;stack level: max 3: 2 by nested T0-irq + rc-irq
interrupt
 if RCCHAN			;if RC-interface:
   ifdef __16F648A
	btfsc	INTCON,INTF	;interrupt from RC?
   else
	btfsc	INTCON,RAIF	;some change on RC-inputs?
   endif
	goto	rc_isr		;y: maybe RC-interrupt	
 endif

;must be timer0 interrupt:
t0_isr
	bcf	INTCON,T0IF	;clr timer int.
	bsf	INTCON,GIE	;allow rc-interrupt

	incfsz	t0intcnt,F	;FF->0 if no rentrance
	goto	t0reent		;reentrance: just reload T0

;-------- process Timer0 jobs all within ISR:
t0irq1
	movwf	save_W		;context saving: W
	swapf	STATUS,W	;STATUS -> W (no change of flags!)
	clrf	STATUS		;select bank 0
	movwf	save_SR		;save STATUS

	movlw	TMR0LOAD
	subwf	TMR0,F		;irq every PWMTU microseconds

pwm_tu
	bsf	INTCON,GIE	;(case looped from skip1ms)

;============= Macro for PWM-processing of one channel in ISR:
procpwmout macro chan,port,bit
    local seton,setcnt,skip
	decfsz	pwmcnt+chan,F	;decrement PWM timer
	goto	skip		;no timeout yet here

	movf	pwmval+chan,W
 if ACTIVELOW
	btfsc	port,bit	;current status:
 else
	btfss	port,bit	;current status:
 endif
	goto	seton		;is off, turn it on
;LED is on, turn off
	sublw	PWMCYC-1	;PWMCYC-1 - pwmval -> W
	btfss	STATUS,C
	goto	skip		;pwmval >= PWMCYC: no "off" phase at all

	addlw	1
 if ACTIVELOW
	bsf	port,bit	;switch LED off
 else
	bcf	port,bit	;switch LED off
 endif
	goto	setcnt

;LED is off, turn LED on:
seton
	btfsc	STATUS,Z
	goto	skip		;no "on" phase at all, skip
 if ACTIVELOW
	bcf	port,bit	;switch LED on
 else
	bsf	port,bit	;switch LED on
 endif
setcnt
	movwf	pwmcnt+chan	;set timer: duration of next phase
skip
	ENDM
;===============================================================
 ifdef __16F648A
;--------- PWM processing, channel 0:
	procpwmout	0,PORTB,4

	if NCHAN > 1	
;--------- PWM processing, channel 1:
	procpwmout	1,PORTB,5
	endif

	if NCHAN > 2	
;--------- PWM processing, channel 2:
	procpwmout	2,PORTB,6
	endif

	if NCHAN > 3	
;--------- PWM processing, channel 3:
	procpwmout	3,PORTB,7
	endif

	if NCHAN > 4	
;--------- PWM processing, channel 4:
	procpwmout	4,PORTA,6
	endif

	if NCHAN > 5	
;--------- PWM processing, channel 5:
	procpwmout	5,PORTA,7
	endif

	if NCHAN > 6	
;--------- PWM processing, channel 6:
	procpwmout	6,PORTA,0
	endif

	if NCHAN > 7	
;--------- PWM processing, channel 7:
	procpwmout	7,PORTA,1
	endif
 else
;--------- PWM processing, channel 0:
	procpwmout	0,PORTC,0

	if NCHAN > 1	
;--------- PWM processing, channel 1:
	procpwmout	1,PORTC,1
	endif

	if NCHAN > 2	
;--------- PWM processing, channel 2:
	procpwmout	2,PORTC,2
	endif

	if NCHAN > 3	
;--------- PWM processing, channel 3:
 ifdef __12F675			
	procpwmout	3,PORTC,4	;12Fxxx: port is GP4
 else
	procpwmout	3,PORTC,3	;16Fxxx: port is RC3
 endif
	endif

	if NCHAN > 4	
;--------- PWM processing, channel 4:
 ifdef __12F675			
	procpwmout	4,PORTC,5	;12Fxxx: port is GP5
 else
	procpwmout	4,PORTC,4	;16Fxxx: port is RC4
 endif
	endif

	if NCHAN > 5	
;--------- PWM processing, channel 5:
	procpwmout	5,PORTC,5
	endif

	if NCHAN > 6	
;--------- PWM processing, channel 6:
	procpwmout	6,PORTA,4
	endif

	if NCHAN > 7	
;--------- PWM processing, channel 7:
	procpwmout	7,PORTA,5
	endif
 endif
;--------- PWM job complete, maintain system timer
	movlw	PWMTU
	addwf	timesum,F	;add time that has passed
	btfss	STATUS,C
	goto	skip1ms
	incfsz	timesum+1,F
	goto	skip1ms

	incf	systicks,F	;1 more tick has passed
	movlw	LOW TICKTIME
	subwf	timesum,F
	btfss	STATUS,C
	decf	timesum+1,F
	movlw	HIGH TICKTIME
	subwf	timesum+1,F

skip1ms
	bcf	INTCON,GIE	;disable T0-int. (critical section)
	decf	t0intcnt,F	
	btfss	t0intcnt,7	;any more timeouts to process?
	goto	pwm_tu		;y: do it one more time...
	
i_exit  
	swapf	save_SR,W
	movwf	STATUS		;restore original STATUS
	swapf	save_W,F	;(keep Z-flag!)
	swapf	save_W,W	;restore original W
	retfie

;T0 interrupt while last T0 int not yet finished:
t0reent
	movwf	save_W2		;context saving: W
	swapf	STATUS,W	;STATUS -> W (no change of flags!)
	clrf	STATUS		;select bank 0
	movwf	save_SR2	;save STATUS

	movlw	TMR0LOAD
	subwf	TMR0,F		;irq every PWMTU microseconds

	swapf	save_SR2,W	
	movwf	STATUS		;restore original STATUS
	swapf	save_W2,F	;(keep Z-flag!)
	swapf	save_W2,W	;restore original W
	retfie
	
;******************************* Lookup-Tables  
;!!! tables must not cross page boundaries !!!

;=========== return bit(channel:1..NCHAN) in W
chanbit
	decf	channel,W

;=============== returns bit value of bit position (W)
ret1248
	andlw	0x07		;(for sure...)
	addwf	PCL,F
	retlw	1
	retlw	2
	retlw	4
	retlw	8
	retlw	16
	retlw	32
	retlw	64
	retlw	128

;================ execution of PCP control instruction (W)
ctrlinstr
	addlw	ctrlt1-ctrlt2	
	btfsc	STATUS,C	;check for valid range of opcodes
	goto	pcp_reset	;invalid opcode or Reset => Reset!
	addlw	ctrlt2-ctrlt1	;restore for table access...
	addwf	PCL,F
ctrlt1
	goto	pcp_smthoff	;0xC0: smooth mode off
	goto	pcp_smthon	;0xC1: smooth mode on
	goto	pcp_haltswoff	;0xC2: halt while switch is off
	goto	pcp_haltswon	;0xC3: halt while switch is on
	goto	pcp_jswoff	;0xC4: jump if channel switch is off
	goto	pcp_jswon	;0xC5: jump if channel switch is on
	goto	pcp_clrsw	;0xC6: clear channel's control switch
	goto	pcp_setsw	;0xC7: set channel's ctrl.switch 
	goto	pcp_jump	;0xC8: JUMP
	goto	pcp_rnddly	;0xC9: random delay
	goto	pcp_rndbrt	;0xCA: random brightness
	goto	pcp_rndjmp	;0xCB: random jump
	goto	pcp_jswxoff	;0xCC: jump if switch (x) is off
	goto	pcp_jswxon	;0xCD: jump if switch (x) is on
	goto	pcp_clrswx	;0xCE: clear switch (x)
	goto	pcp_setswx	;0xCF: set switch (x)
	goto	pcp_nop2	;0xD0: just header! Don't process here
	goto	pcp_nop2	;0xD1: just header! Don't process here
	goto	pcp_nop2	;0xD2: just header! Don't process here
	goto	pcp_nop2	;0xD3: just header! Don't process here
	goto	pcp_nop2	;0xD4: just header! Don't process here
	goto	pcp_nop2	;0xD5: just header! Don't process here
	goto	pcp_nop2	;0xD6: just header! Don't process here
	goto	pcp_nop2	;0xD7: just header! Don't process here
	goto	pcp_jvarb	;0xD8: jump if var below
	goto	pcp_jvarae	;0xD9: jump if var above or equal
	goto	pcp_jvarxb	;0xDA: jump if var x below
	goto	pcp_jvarxae	;0xDB: jump if var x above or equal	
	goto	pcp_dly		;0xDC: delay
ctrlt2

;****************** Device initialization
init
 ifdef __16F636
	movlw	0x07
	movwf	CMCON0		;comparator off
 else
	movlw	0x07
	movwf	CMCON
 endif
	bsf	STATUS,RP0	;select Register-bank 1
 errorlevel-302
;--------------------------------------------------------------
 ifdef __16F648A		;MPU = 16F627/628/648:
trisval	set	~AUXOUTA
   if NCHAN >4
trisval &=	B'10111111'	;RA6 = output #4
   endif
   if NCHAN >5
trisval &=	B'01111111'	;RA7 = output #5
   endif
   if NCHAN >6
trisval &=	B'11111110'	;RA0 = output #6
   endif
   if NCHAN >7
trisval &=	B'11111101'	;RA1 = output #7
   endif
	movlw	trisval
	movwf	TRISA

trisval	set	~AUXOUTB & B'11101111'	;RB4 = output #0
   if NCHAN >1
trisval &=	B'11001111'	;RB5 = output #1
   endif
   if NCHAN >2
trisval &=	B'10001111'	;RB6 = output #2
   endif
   if NCHAN >2
trisval &=	B'00001111'	;RB7 = output #3
   endif
   if RCCHAN
trisval |=	B'00000001'	;RB0 = input(RC pulse)
   endif
 	movlw	trisval
	movwf	TRISB
;--------------------------------------------------------------
 else				;MPU = 16F630/16F636:
   ifndef __12F675
trisval	set	~AUXOUTA
     if RCCHAN == 2
trisval |=	1		;RA0 = input (RC #2)
     endif
     if NCHAN > 6
trisval &=	B'101111'	;RA4 = output #7
     endif
     if NCHAN > 7	
trisval &=	B'011111'	;RA5 = output #8
     endif
	movlw	trisval
	movwf	TRISA

trisval	set	~AUXOUTC
trisval &= (0xFF << NCHAN)
	movlw	low trisval
	movwf	TRISC
;--------------------------------------------------------------
   else				;MPU = 12F629/12F675:
trisval	set	~AUXOUTC
     if NCHAN <4
trisval &= (0xFF << NCHAN)	;GP0/GP1/GP2 = first 3 outputs
     else
trisval &= (0xFF << NCHAN+1)	;skip GP3 (=always input)
     endif
     if RCCHAN == 2
trisval |=	B'100000'	;GP5 = input (RC #2)
     endif
	movlw	low trisval
	movwf	TRISIO
   endif
 endif

 ifdef __16F636
  if CLOCK == 8
	bsf	OSCCON,4	;switch to 8 MHz clock
  endif
 endif
	clrwdt
	movlw	B'01011000'	;internal Pull-Ups enabled,
	movwf	OPTION_REG	;Prescaler->WDT, T0-clock=instr.-cycle
 ifdef __12F675
	clrf	ANSEL
	call	0x3FF
	movwf	OSCCAL
	bsf	IOC,3		;interrupt on change for RC 
 else
   ifndef __16F648A
	bsf	IOCA,3		;interrupt on change for RC channel 1
   endif
 endif
	bcf	STATUS,RP0	;back to Bank 0
 errorlevel +302
	
	movlw	0x20
	movwf	FSR
clrloop
	clrf	INDF		;initialize RAM with 00s
	incf	FSR,F
	btfss	FSR,7
	goto	clrloop

;	decf	ee_edit,F	;= 0xFF: edit not active
	decf	timesum+1,F	;= 0xFF
	decf	t0intcnt,F	;= 0xFF
	incf	tc_100,F	;= 1

 if RCCHAN
	movlw	STARTSW
	movwf	memswtch
	decf	rcval,F		;-1: not yet received any valid RC-pulse
   if RCCHAN>1
	decf	rcval+1,F
   endif
 endif

	movlw	NCHAN
	movwf	channel		;initialize PCP pointers
inipcplp
	call	pcpip_fsr	;pcpip(channel) -> FSR 
	decf	channel,W
	call	ee_read		;get PCP startaddress(channel)
	movwf	INDF		;write pcpip(channel)

 if COMMTICK == 0
	movlw	systcnt-1
	call	chan_fsr
	incf	INDF,F		;= 1: immediate time-out
 endif
	call	tickcnt_fsr
	incf	INDF,F		;=1...

	decfsz	channel,F
	goto	inipcplp

	bsf	T1CON,T1CKPS1	;T1 prescaler: T=4s
 if CLOCK == 8
	bsf	T1CON,T1CKPS0
 endif
 ifdef __16F648A
   if RCCHAN
	movlw	B'10110000'	;enable T0-int and RC-interrupt
   else
	movlw	B'10100000'	;enable T0-int
   endif
 else
   if RCCHAN
	movlw	B'10101000'	;enable T0-int and RA-interrupt
   else
	movlw	B'10100000'	;enable T0-int
   endif
 endif
	movwf	INTCON	

	clrf	switch	

;__________________________________________________________________

mainlp
 if RCCHAN			;case RC interface:
   ifdef __16F648A
	btfss	INTCON,INTE	;got an RC-pulse?
   else
	btfss	INTCON,RAIE	;got an RC-pulse?
   endif
	call	proc_rc		;y: do rc-pulse processing

;	btfsc	flags,RCFLG1
;	call	new_rcval1	;new valid RC-value(1)
;   if RCCHAN > 1
;	btfsc	flags,RCFLG2
;	call	new_rcval2	;new valid RC-value(2)
;   endif
 endif 
	movf	systicks,W
	btfsc	STATUS,Z	;systicks > 0?
	goto	mainlp		;no, nothing to execute

;time went forward by 1 systick:
	bcf	flags,SKIPSMTH
	decfsz	tc_100,F
	goto	mlp2
	
	movlw	100000/TICKTIME
	movwf	tc_100
;	incf	time10,F	;1/10 second over
 if RCCHAN
	call	rcserv10	;call rc-service routine (memory switch)
 endif

 if COMMTICK	;--- common tick-time:
mlp2
	decfsz	systicks,F	;1->0 if right in time
	bsf	flags,SKIPSMTH	;late: skip smooth-processing to speed up
	movlw	NCHAN
	movwf	channel		;process systick for each channel

mlpch
	call	chantick	;process tick(channel)

 else	;--- individual tick times per channel:
mlp2
	decf	systicks,F
	addlw	-10		;more than 10 ms late?
	btfsc	STATUS,C
	bsf	flags,SKIPSMTH	;late: skip smooth-processing to speed up

	movlw	NCHAN
	movwf	channel		;process systick for each channel

	addlw	systcnt
	movwf	FSR		;address: systcnt+NCHAN
mlpch
	decf	FSR,F
	decfsz	INDF,F		;decr. systick-counter(channel)
	goto	mlpchnxt	;no timeout yet

	movlw	0xD0		;(opcode: DEFTICKLEN)
	call	getparm		;get length of tick (channel)
	btfss	STATUS,C
	movlw	10		;parameter not present: default=10ms
	movwf	INDF		;reload systcnt(channel)

	call	chantick	;process tick(channel)
	movf	channel,W
	addlw	systcnt-1
	movwf	FSR		;restore FSR to tick-counter
mlpchnxt
 endif

	decfsz	channel,F	;next channel...
	goto	mlpch
	goto	mainlp

;================= Step PCP(channel) forward by one tick
chantick
	call	tickcnt_fsr	;access tickcnt(channel)
	decfsz	INDF,F		;decrement by 1
	goto	chantck_wt	;time not yet over

;	bcf	flags,JUMPFLG

;timeout of tick-counter: execute next program step or steps
chantout
	call	read_ip
	movwf	y		;instruction or brightness val. from EEPROM -> y
	addlw	0x40
	btfsc	STATUS,C
	goto	chanctrl	;>= 0xC0: interpret control instr.

	call	nextbrght	;opcode <0xC0: next brightness value
	goto	chantck1	

;interpret control instruction:	
chanctrl
	call	ctrlinstr

chantck1
	call	tickcnt_fsr	;access delay(channel)
	movf	INDF,F		; == 0 ?
	btfsc	STATUS,Z
	goto	chantout	;no delay: next instruction immediately
	return			;>0: processing finished for this step


;set new brightness value from PCP:
nextbrght
	call	setpwm		;new pwm value
	call	read_ip		;read delay value from EEPROM
	movwf	temp
	call	tickcnt_fsr	;select tick-counter(channel)
	movf	temp,W
	movwf	INDF		;save new delay time
	return

;still waiting for timeout: process brightness-interpolation
chantck_wt
	btfsc	flags,SKIPSMTH	;discard smooth-processing?
	return			;y (we are late - this will speed up tick processing)

	movf	INDF,W		;remaining ticks till next operation
	movwf	x1		;-> x1
	call 	chanbit		;W = bit(channel)
	andwf	smooth,W	;smooth mode?
	btfsc	STATUS,Z
	return			;no, task finished

	call	pcpip_fsr
	movf	INDF,W		;IP -> W
	call	ee_read		;get next opcode
	movwf	y		;brightness at end of current interval
	addlw	0x40
	btfsc	STATUS,C
	return			;ctrl.-instruction, no interpolation here

	movlw	PWMCYC
	movwf	x
	call	multiply	;adjust to PWM value -> y
	movf	y,W
	movwf	y2		;pwmval. at end of interval -> y2

 errorlevel-302
	bsf	STATUS,RP0	;select bank 1
	decf	EEADR,W		;access previous byte (= ticks)
	call	ee_read
	movwf	x2		;full interval (ticks)

	bsf	STATUS,RP0	;select bank 1
	decf	EEADR,W		;access previous byte (= intensity)
 errorlevel+302
	call	ee_read
	movwf	y
	addlw	0x40
	btfsc	STATUS,C
	return			;previous opcode = control instr. - no interpolation

	movlw	PWMCYC
	movwf	x
	call	multiply	;adjust to PWM value -> y
	movf	y,W
	movwf	y1		;PWM val. at start of interval -> y1

	movf	x1,W		;remaining ticks
	subwf	x2,W		;full interval - remaining ticks -> x
	movwf	x			 
	call	interpolate
	goto	setpwmadj	;adjust and store PWM value

;================= sets pwm of (channel) from brightness value (y)
;input value of y=0x80 means 100%, will then be adjusted by
;brightness adjustment value from PCP-header (0x80 means * 1.0)   
;if brightness changes by high value: forces fast adjustment of PWM.
setpwm
	movlw	PWMCYC
	movwf	x
	call	multiply	;adjust to PWM value

setpwmadj			;adjust by adj.value for this channel:
	movlw	0xD1		;opcode: DEFBRIGHT
	call	getparm
	movwf	x		;(0x80 means *1.0)
	btfsc	STATUS,C	;case no such data in header: *1.0 -> skip multiply
	call	multiply	;new PWM-value = y

	movlw	pwmval-1
	addwf	channel,W
	movwf	FSR
	movf	y,W
	subwf	INDF,W		;W= difference from old level
	btfss	STATUS,C
	xorlw	0xff		;negative, change sign
	sublw	PWMCYC/2	;C=0 for bigger changes (> half of range)
	rrf	x,F		;save C -> MSB(x)
	movf	y,W
	movwf	INDF		;store new value

	movlw	pwmcnt-1
	addwf	channel,W
	movwf	FSR
	movf	INDF,W		;time till next change
	sublw	PWMCYC		;>PWMCYC? (maybe if full on/off!)
	btfss	STATUS,C
	goto	clrpwmcnt	;y, force immediate PWM change

	btfsc	x,7		;big change of value?
	return			;n: let PWM adjust at next regular cycle

clrpwmcnt
	movlw	1
	movwf	INDF		;force start of new pwm-cycle on next pwm-irq
	return


;================ Linear Interpolation
;computes Y-value for given X-val. 
;input:
;y1 = Y-value for X=0
;y2 = Y-value for X=x2
;x2 = X-value: length of interval
;x = x input value (1 ... x2-1)
;output:
;y = Y-value for given x

interpolate
	clrf	x1
	btfsc	x2,7
	goto	interp_loop	;x2 >= 0x80

	bcf	STATUS,C
interp_max				
	rlf	x2,F		;maximize for higher precision
	rlf	x,F
	btfss	x2,7
	goto 	interp_max

interp_loop
	movf	y1,W
	addwf	y2,W
	movwf	y
	rrf	y,F		;= mean val. of y1, y2

	movf	x1,W
	addwf	x2,W
	movwf	temp		
	rrf	temp,F		;= mean val. of x1, x2
	movf	temp,W
	subwf	x,W
	btfsc	STATUS,Z
	return			;exact match -> iteration complete

	movf	y,W
	xorwf	y1,W
	btfsc	STATUS,Z	;y == y1?
	goto	interp_exit	;Y: result is y1 or y2

	movf	y,W
	xorwf	y2,W
	btfsc	STATUS,Z	;y == y2?
	goto	interp_exit	;Y: result is y1 or y2
 
	movf	temp,W		;mean(x1,x2) -> W
	btfss	STATUS,C
	goto	interp_less	;C=0 => x < mean(x1,x2) 
;x > mean(x1,x2):
	movwf	x1		;new interval = mean..x2
	movf	y,W
	movwf	y1		;new y within y..y2
	goto	interp_loop

;x < mean(x1,x2):
interp_less
	movwf	x2		;new interval = x1..mean
	movf	y,W
	movwf	y2		;new y within y1..y
	goto	interp_loop

;difference(y1,y2) <= 1, use y1 or y2 depending on x < or > mean(x1,x2)
interp_exit
	movf	y1,W
	btfsc	STATUS,C
	movf	y2,W		;x > mean(x1,x2)
	movwf	y
	return

;===================== Multiplication subroutine
;multiplies y * x -> y
;x, y = fixpoint, 0x80 means 1.0

multiply
	movlw	8
	movwf	temp		;loop counter
	movf	y,W
	clrf	y

mul_lp
	rrf	x,F		;next bit of multiplier -> C
	btfsc	STATUS,C
	addwf	y,F		;bit was 1: add multiplicant to result
	rrf	y,F		;result /2
	decfsz	temp,F
	goto	mul_lp		;next loop

	rlf	y,F
	btfss	STATUS,C
	return
	movlw	0xFF		;overflow: set result to max.value
	movwf	y
	return

;=================== Read byte (pcpip(channel)) from EEPROM
;with increment of IP
read_ip
	call	pcpip_fsr
	movf	INDF,W		;IP -> W
	incf	INDF,F		;increment IP 

;================ Read byte from EEPROM
;input: W = address
;output: EEPROM data -> W

ee_read
 errorlevel-302
	bsf	STATUS,RP0	;select bank 1
	movwf	EEADR
;	subwf	ee_edit,W	;address selected for editing?
;	btfsc	STATUS,Z
;	goto	ee_redit	;y: read from edit-register rather than EPROM

	bsf	EECON1,RD	;initiate read op.
	movf	EEDATA,W	;get data
	bcf	STATUS,RP0	;select standard bank (0)
 errorlevel+302
	return

;ee_redit
;	movf	ee_edit+1,W
;	bcf	STATUS,RP0
;	return

;================== get parameter from PCP header(channel)
;input: W = opcode for selection of parameter
;output W = parameter value, C=1
;	or C=0 if requested parameter was not found in header data
getparm
 errorlevel-302
	movwf	temp		;save opcode to search for
	decf	channel,W
	bsf	STATUS,RP0	;bank 1 for EEPROM access
	movwf	EEADR
	bsf	EECON1,RD	
	movf	EEDATA,W		
	movwf	EEADR		;PCP start addr. -> EEADR

getparmlp
	bsf	EECON1,RD
	incf	EEADR,F		;inc addr -> par.value
	movf	temp,W
	subwf	EEDATA,W	;does opcode match?
	btfss	STATUS,Z
	goto	getprmnext	;no match	

	bsf	EECON1,RD	;right opcode: read parameter value
	movf	EEDATA,W	;-> W
	bcf	STATUS,RP0	;select standard bank (0)
	return			;(C=1 by previous subwf)
			
getprmnext
	incf	EEADR,F		;set ptr: next opcode
	bsf	EECON1,RD
	movlw	0xD0		;(DEFTICKLEN) = lowest opcode for header data
	subwf	EEDATA,W	;current opcode - DEFTICKLEN
	btfss	STATUS,C
	goto	gotnoparm	;below DEFTICKLEN: reached end of header

	addlw	-4		;only opc. 0xD0..0xD3 are header
	btfss	STATUS,C
	goto	getparmlp	;opcode was still header

gotnoparm
	bcf	STATUS,C	;header does not contain requested parameter
	bcf	STATUS,RP0
	return 
 errorlevel+302

;================== set FSR to tickcnt(channel)
tickcnt_fsr
	movlw	tickcnt-1
	goto	chan_fsr

;================== set FSR to pcp_ip(channel)
pcpip_fsr
	movlw	pcpip-1

;================== W+channel -> FSR
chan_fsr
	addwf	channel,W
	movwf	FSR
	return

;********************************************************
;**** Subroutines for PCP control code processing    ****

;__________________________________________________________
pcp_reset			;P_RESET: Reset pcp(channel) to startadr,
;	btfsc	flags,JUMPFLG	;if there was a jump already in same tick:
	call	inc_tickcnt	;avoid deadlock: further ops on next tick	

;	bsf	flags,JUMPFLG	;it's like a jump...
	call	pcpip_fsr	;set FSR = ip(channel)
	decf	channel,W
	call	ee_read		;read pcp startadr.(channel)
	movwf	INDF		;-> ip

	clrf	y
	call	setpwm		;set output level = off

;__________________________________________________________
pcp_smthoff			;SMOOTHOFF: smooth mode off
	call	chanbit
	xorlw	0xFF
	andwf	smooth,F
	return
;__________________________________________________________
pcp_smthon			;SMOOTHON: smooth mode on
	call	chanbit
	iorwf	smooth,F
	return
;__________________________________________________________
pcp_jvarxb			;jump if variable value < arg
	call	var_sel
jvarb1
	btfsc	STATUS,C
	goto	pcp_skip2	;invalid var-adress: skip value, no jump

	movwf	temp		;cont. of variable
	call	read_ip		;get argument = value to compare with
	subwf	temp,W		;var-arg: C=0 if var<arg
	btfsc	STATUS,C
	goto	pcp_skip	;arg >= var: no jump, skip
	goto	pcp_jump	;do the jump
;__________________________________________________________
pcp_jvarxae			;jump if variable value >= arg
	call	var_sel
jvarae1
	btfsc	STATUS,C
	goto	pcp_skip2	;invalid var-adress: skip value, no jump

	movwf	temp		;cont. of variable
	call	read_ip		;get argument = value to compare with
	subwf	temp,W		;var-arg: C=0 if var<arg
	btfsc	STATUS,C
	goto	pcp_jump	;do the jump
	goto	pcp_skip	;arg < var: no jump, skip
;__________________________________________________________
pcp_jvarb			;jump if chennel's var <...
	call	dvar_sel
	goto	jvarb1
;__________________________________________________________
pcp_jvarae			;jump if channel's var >= ...
	call	dvar_sel
	goto	jvarae1

;======= skip 2 bytes:
pcp_skip2
	call	read_ip
;__________________________________________________________
;2 byte "no operation":
pcp_nop2		
	incf	INDF,F
	return
;__________________________________________________________
pcp_haltswon			;halt while dedicated switch is on	
	call	dsw_sel		;select bit
	btfsc	STATUS,C
	return			;invalid switch: no halt

	andwf	INDF,W
	btfsc	STATUS,Z
	return			;bit is clr: no halt
	goto	halt1
;__________________________________________________________
pcp_haltswoff			;halt while dedicated switch is off	
	call	dsw_sel		;select bit
	btfsc	STATUS,C
	return			;invalid switch: no halt

	andwf	INDF,W
	btfss	STATUS,Z
	return			;bit is set: no halt
halt1
	call	pcpip_fsr	;FSR = instruction pointer
	decf	INDF,F		;restore IP to "halt" instr.

inc_tickcnt
	call	tickcnt_fsr
	incf	INDF,F		;= 1: wait one tick before next action

 if COMMTICK==0			;individual tick times (probably very long!):
	movlw	systcnt-1
	call	chan_fsr
	movlw	10		;10 ms until next operation, 
	movwf	INDF		;regardless of ticktime-setting!
 endif
	return
;__________________________________________________________
pcp_jump			;JUMP
;	btfsc	flags,JUMPFLG	;if there was a jump already in same tick:
	call	inc_tickcnt	;avoid deadlock: continue on next tick	
	
;	bsf	flags,JUMPFLG	;set flag: "some jump was executed"
	call	read_ip		;get destination for jump
	movwf	INDF		;-> new IP
	return
;__________________________________________________________
pcp_rndjmp			;random jump
	call	rnd8
	andlw	0x7F
	movwf	temp		;save 7 bit random number
	call	read_ip		;get "chance"
	subwf	temp,W
	btfss	STATUS,C
	goto	pcp_jump	;rnd7 < chance: execute jump
	goto	pcp_skip
;__________________________________________________________
pcp_jswxoff			;JSWXOFF: jump if switch (X) is off
	call	switch_sel	;select bit
	goto	jswoff1		;jump or not to jump...
;__________________________________________________________
pcp_jswoff			;JSWOFF: jump if channel switch is off
	call	dsw_sel		;select bit: dedicated channel switch
jswoff1
	btfsc	STATUS,C
	goto	pcp_jump	;C=1: invalid switch#, treat as "off"!

	andwf	INDF,W
	btfsc	STATUS,Z
	goto	pcp_jump	;switch is off: execute jump

pcp_skip			;condition = false: don't execute jump
	call	pcpip_fsr
	incf	INDF,F		;skip jmp address-byte
	return	
;__________________________________________________________
pcp_jswxon			;JSWXON: jump if switch (X) is on
	call	switch_sel	;select bit
	goto	jswon1		;jump or not to jump...
;__________________________________________________________
pcp_jswon			;JSWON: jump if ch.-switch is on
	call	dsw_sel		;select bit: dedicated channel switch
jswon1
	btfsc	STATUS,C
	goto	pcp_skip	;C=1: invalid switch#!

	andwf	INDF,W	
	btfss	STATUS,Z
	goto	pcp_jump	;switch is on: execute jump
	goto	pcp_skip	;off: skip
;__________________________________________________________
pcp_clrswx			;clear switch X:
	call	switch_sel	;get access to that switch
	goto	clrsw1
;__________________________________________________________
pcp_clrsw			;clear dedicated control switch:
	call	dsw_sel		;select bit
clrsw1
	btfsc	STATUS,C
	return			;C=1: invalid switch#!

	xorlw	0xFF
	andwf	INDF,F		;reset bit
	return
;__________________________________________________________
pcp_setswx			;set switch X:
	call	switch_sel	;get access to that switch
	goto	setsw1
;__________________________________________________________
pcp_setsw			;set dedicated control switch:
	call	dsw_sel		;select bit
setsw1
	btfsc	STATUS,C
	return			;C=1: invalid switch#!

	iorwf	INDF,F		;set bit
	return
	

;================= select dedicated switch (channel)
;returned: FSR = RAM address, W = bitmask
;C=1 if address is invalid
dsw_sel
	movlw	0xD2		;opcode: DEFSWITCH
	call	getparm		;switch defined in header?
	btfss	STATUS,C	;C=1: yes, W = switch#
	decf	channel,W	;C=0: no, use channel#
	goto	switch_s1

;================= select switch (read_ip)
;returned: FSR = address of register, W = bitmask
;C=1 if address is invalid
;
;map of switches:
;0..7 = default channel dedicated switches
;8..11 = RC memory switches, RC-channel 1
;12..15 = RC memory switches, RC-channel 2
;16..23 = I/O port A
;24..31 = I/O port B, C or GPIO, respectively
;32..47 = (reserved: RC Multiswitch decoder)
;48: RC input pulse detection (1: no valid RC pulses)

switch_sel
	call	read_ip	
switch_s1
	movwf	temp
	addlw	-16		;number < 16?
	btfsc	STATUS,C
	goto	switch_port	;>= 16: i/o-port access

	movlw	switch		;0..7: select default switches
	btfsc	temp,3
	movlw	memswtch	;>= 8: select rc memory switch
	goto	sw_fsr

switch_port
	addlw	-16
 if RCCHAN
	btfss	STATUS,C
	goto	range_port_sw	;<32: in range of port lines

	addlw	-16
	btfss	STATUS,C
	goto	range_ms	;<48: in range of multiswitch

	btfss	STATUS,Z
	return			;>48: invalid switch number (C=1)
	
	bcf	STATUS,C	;== 48: rc pulse detection
	movlw	rcdetect	;>=0 if valid rc pulses detected
	movwf	FSR
	retlw	B'10000000'	;mask MSB 	
	
 else
	btfsc	STATUS,C
	return			;>=32: invalid (C=1)!
 endif
range_port_sw
 ifdef __16F648A
	movlw	PORTB
 else
	movlw	PORTC		;= GPIO on 12F675
 endif
 ifdef __12F675
	btfsc	temp,3	
	goto	sw_fsr		;ok, 24..31 = GPIO
	bsf	STATUS,C	;invalid switch number!
	return
 else
	btfss	temp,3
	movlw	PORTA		;16..23: port A
 endif
sw_fsr	movwf	FSR
	movf	temp,W
	goto	ret1248		;return mask-bit (C=0)

;--------- switch number = 32..47, e.g. Multiswitch
range_ms
	bsf	STATUS,C	;not implemented here!
	return

;================= select channel's dedicated variable
dvar_sel
	movlw	0xD3		;opcode: DEFVAR
	call	getparm		;var defined in header?
	btfss	STATUS,C	;C=1: yes, W = var#
	decf	channel,W	;C=0: no, use channel#
	goto	var_sel1

;================= select variable (read_ip)
;returned: FSR = address of register, W=data read from register
;C=1 if address is invalid
;
;map of registers:
;0: data of RC-channel 1
;1: data of RC-channel 2

var_sel
	call	read_ip
var_sel1
	bsf	STATUS,C	
 if RCCHAN			;if some RC-input is implemented:
	movwf	FSR
   if RCCHAN>1
	andlw	0xFE		;2 RC inputs: 0/1 allowed!
   else
	andlw	0xFF		;only 1 RC-input: must be 0
   endif	
	btfss	STATUS,Z
	return			;invalid variable number! 

	movlw	rcval
	addwf	FSR,F
	movf	INDF,W
 endif
	return


;__________________________________________________________
pcp_dly				;Delay:
	call	read_ip		;get delay value
	movwf	y
	call	tickcnt_fsr
	movf	y,W
	movwf	INDF		;set new delay time
	return

;======================= generate next 8 bit random number
rnd8
	bcf	STATUS,C
	rrf	random,W
	btfsc	STATUS,C
	xorlw	0xB4
	addwf	pwmcnt,W
	movwf	random
	return

;__________________________________________________________
pcp_rnddly			;Random Delay:
	call	read_ip		;get val for max. delay
	movwf	x
	call	rnd8		;new 8 bit random value
	addlw	1
	movwf	y
	rrf	y,F		;y = random number 0.0 ... 1.0
	call	multiply
	call	tickcnt_fsr
	movf	y,W
	movwf	INDF		;set new delay time
	return
;__________________________________________________________
pcp_rndbrt			;set Random Brightness:
	call	rnd8
	andlw	0x7F
	movwf	x		;7 bit random value

	call	read_ip		;get min. brightness
	movwf	temp
	call	read_ip		;get max. brightness 
	movwf	y

	movlw	pwmval-1
	addwf	channel,W
	movwf	FSR
	movf	temp,W
	movwf	INDF		;min. brightness
	subwf	y,F		;difference (max. - min.) -> y
	call	multiply	;y = diff. * rnd7
	movf	INDF,W		;min.
	addwf	y,F		;y += min.
	goto	nextbrght	;sets brght=Y, next delay
;__________________________________________________________

